***********************************************************************
Example:

Here's an example of an interactive S session that uses these
functions. The dataset is the same one used by Little and
Schlucter (1985), in the ASCII file called "stlouis.dat". 

# Read in the data, name the variables, and put the 3 categorical
# variables in the leftmost columns.
x_scan("c:\\mix\\stlouis.dat")
x_matrix(x,69,8,T)
tmp_format(x[,1])
x_x[,2:8]
dimnames(x)_list(tmp,c("G","R1","V1","D1","R2","V2","D2"))
x_x[,c(1,4,7,2,3,5,6)]
x[x==999]_NA
rm(tmp)
# Perform preliminary manipulations on x. The categorical 
# variables need to be coded as consecutive positive integers
# beginning with 1. 
s_prelim.mix(x,3)
# look at missingness patterns
print(s$r)
# Try EM for general location model without restrictions. This
# algorithm converges after 168 iterations.
thetahat1_em.mix(s)
# look at the parameter estimates and loglikelihood
print(getparam.mix(s,thetahat1))
print(getparam.mix(s,thetahat1,corr=T))
print(loglik.mix(s,thetahat1))
# take 100 steps of data augmentation starting from thetahat1
rngseed(1234567)
newtheta_da.mix(s,thetahat1,steps=100,showits=T)
# re-run em beginning from newtheta; should converge after 86 
# iterations.
thetahat2_em.mix(s,newtheta)
# Notice that the loglikelihood at thetahat2 is somewhat different
# than at thetahat1. Examination of thetahat1 and thetahat2 reveals
# that EM has converged to different values. The likelihood is
# multimodal.
print(loglik.mix(s,thetahat2))
print(getparam.mix(s,thetahat2))
# Now try fitting a model with restrictions. We'll first fit the 
# "null model" described in chap. 9 of Schafer (1996), which 
# fits the margins G and D1xD2 in the contingency table, and
# has a full D1*D2 interaction for each continuous variable
# but no effect for G.
margins_c(1,0,2,3)
intercept_rep(1,12)
d1_c(-1,-1,-1,1,1,1,-1,-1,-1,1,1,1) 
d2_c(-1,-1,-1,-1,-1,-1,1,1,1,1,1,1)
design_cbind(intercept,d1,d2,d1d2=d1*d2)
rm(intercept,d1,d2)
thetahat3_ecm.mix(s,margins,design)
print(loglik.mix(s,thetahat3))
# If we play around with starting values, we'll find that the
# likelihood under the "null model" is also multimodal.
# Now let's fit the "alternative model".
margins_c(1,2,0,2,3,0,1,3)
glin_c(-1,0,1,-1,0,1,-1,0,1,-1,0,1)
design_cbind(design,glin)
thetahat4_ecm.mix(s,margins,design)
print(loglik.mix(s,thetahat4))
# Now try some imputations. The following commands produce three
# multiple imputations under the alternative model. The imputations
# are proper if we can assume that the data augmentation procedure
# achieves stationarity by 100 steps.
rngseed(454545)
newtheta_dabipf.mix(s,margins,design,thetahat4,steps=100,showits=T)
imp1_imp.mix(s,newtheta,x)
newtheta_dabipf.mix(s,margins,design,newtheta,steps=100,showits=T)
imp2_imp.mix(s,newtheta,x)
newtheta_dabipf.mix(s,margins,design,newtheta,steps=100,showits=T)
imp3_imp.mix(s,newtheta,x)
***********************************************************************
References:

Little, R.J.A.  and Schluchter, M.D. (1985), "Maximum-likelihood
estimation for mixed continuous and categorical data with missing
values." Biometrika, 72, 492-512.

Schafer, J.L. (1997), Analysis of Incomplete Multivariate Data,
Chapman and Hall, London.
***********************************************************************
