".code.to.na"<-
function(x, mvcode)
{
  x[x == mvcode] <- NA
  x
}
".na.to.icode"<-
function(x, mvcode)
{
  x[is.na(x)] <- as.integer(mvcode)
  x
}
".na.to.snglcode"<-
function(x, mvcode)
{
  x[is.na(x)] <- as.single(mvcode)
  x
}

"bipf"<-
function(table, margins, prior = 0.5, start, steps = 1, showits = F)
{
  a <- attributes(table)
  d <- dim(table)
  n <- length(table)
  j <- as.integer(cumprod(d)/d)
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  w <- !is.na(prior)
  if(!missing(start)) {
    if(length(start) != n)
      stop("Starting value has incorrect length")
    if(any(start[w] == 0))
      stop("Starting value on the boundary")
    if(any(!w)) {
      if(any(start[!w] != 0)) {
	stop("Starting value has nonzero elements for structural zeroes")
      }
    }
  }
  if(missing(start)) {
    start <- rep(1, n)
    start[!w] <- 0
    start[w] <- start[w]/sum(start[w])
  }
  prior[!w] <- -999
  if(showits)
    cat(paste("Cycles of Bayesian IPF:", "\n", sep = ""))
  for(it in 1:steps) {
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    tmp <- .Fortran("bipf",
		    n,
		    as.double(table),
		    as.double(start),
		    as.double(prior),
		    length(margins),
		    as.integer(margins),
		    length(d),
		    d,
		    j,
		    d,
		    d,
		    d,
		    as.integer(0))
    if(tmp[[13]] == 1)
      stop("Improper posterior - Check prior.")
    start <- tmp[[3]]
  }
  if(showits)
    cat("\n")
  attributes(start) <- a
  start
}
"da.cat"<-
function(s, start, prior = 0.5, steps = 1, showits = F)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  w <- !is.na(prior)
  if(any(start[w] == 0)) {
    warning("Starting value on the boundary")
  }
  if(any(!w)) {
    if(any(start[!w] != 0)) {
      stop("Starting value has nonzero elements for structural zeroes")
    }
  }
  start <- as.double(start)
  prior <- as.double(prior)
  if(showits)
    cat(paste("Steps of Data Augmentation:", "\n"))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    start <- .Fortran("istepc",
		      s$ncells,
		      start,
		      start,
		      s$npatt,
		      s$p,
		      s$r,
		      s$mdpgrp,
		      s$ngrp,
		      s$mobs,
		      s$nmobs,
		      s$d,
		      s$jmp,
		      as.integer(0),
		      s$d,
		      s$d)
    if(start[[13]] == 1)
      stop("Bad parameter value: assigns zero prob. to an observed event")
      else start <- start[[3]]
    start[w] <- start[w] + prior[w]
    start[!w] <- -999
    start <- .Fortran("pstep1c",
		      s$ncells,
		      start,
		      start,
		      as.integer(0))
    if(start[[4]] == 1)
      stop("Improper posterior - check prior.")
    start <- start[[3]]
  }
  if(showits)
    cat("\n")
  array(start, s$d)
}
"da.mix"<-
function(s, start, steps = 1, prior = 0.5, showits = F)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  prior <- as.double(prior)
  w <- !is.na(prior)
  prior[!w] <- -999
  s$z <- .na.to.snglcode(s$z, 999)
  s$w <- .na.to.icode(s$w, 999)
  tp <- as.integer(s$p)
  tq <- integer(s$q)
  tmp <- .Fortran("tobsm",
		  s$q,
		  s$psi,
		  s$npsi,
		  numeric(s$npsi),
		  s$ncells,
		  matrix(0, s$q, s$ncells),
		  numeric(s$ncells),
		  s$npattz,
		  s$rz,
		  s$mdpzgrp,
		  s$npattw,
		  s$p,
		  s$rw,
		  s$mdpwgrp,
		  s$ngrp,
		  s$mobs,
		  s$mobsst,
		  s$nmobs,
		  s$n,
		  s$z,
		  tp,
		  tq)
  kn1 <- tmp[[4]]
  kn2 <- tmp[[6]]
  kn3 <- tmp[[7]]
  sigma <- start$sigma
  pii <- start$pi
  mu <- start$mu
  t1 <- sigma
  t2 <- mu
  t3 <- pii
  if(showits)
    cat(paste("Steps of Data Augmentation:", "\n"))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    tmp <- .Fortran("istepm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    sigma,
		    mu,
		    pii,
		    kn1,
		    kn2,
		    kn3,
		    t1 = t1,
		    t2 = t2,
		    t3 = t3,
		    s$npattz,
		    s$rz,
		    tq,
		    tq,
		    s$mdpzgrp,
		    s$npattw,
		    s$p,
		    s$rw,
		    tp,
		    s$mdpwgrp,
		    s$ngrp,
		    s$mobs,
		    s$mobsst,
		    s$nmobs,
		    s$n,
		    s$z,
		    s$d,
		    s$jmp,
		    tp,
		    kn3,
		    kn1,
		    s$w,
		    numeric(s$q))
    t1 <- tmp$t1
    t2 <- tmp$t2
    t3 <- tmp$t3
    tmp <- .Fortran("pstepm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    sigma = t1,
		    mu = t2,
		    pi = t3,
		    s$n,
		    s$p,
		    as.double(prior),
		    numeric(s$npsi),
		    matrix(0, s$q, s$q),
		    numeric(s$q),
		    tq,
		    err = numeric(1))
    {
      if(tmp$err == 1)
	stop("Improper posterior--empty cells"
	     )
	else {
	  sigma <- tmp$sigma
	  mu <- tmp$mu
	  pii <- tmp$pi
	}
    }
  }
  if(showits)
    cat("\n")
  mu[, !w] <- NA
  list(sigma = sigma, mu = mu, pi = pii)
}
"da.norm"<-
function(s, start, prior, steps = 1, showits = F, return.ymis = F)
{
  s$x <- .na.to.snglcode(s$x, 999)
  tmpi <- as.integer(numeric(s$p))
  tmpr <- as.single((numeric(s$p)))
  tobs <- .Fortran("tobsn",
		   s$d,
		   numeric(s$d),
		   s$p,
		   s$psi,
		   s$n,
		   s$x,
		   s$npatt,
		   s$r,
		   s$mdpst,
		   s$nmdp,
		   tmpi)[[2]]
  if(missing(prior)) {
    tau <- 0
    m <- -1
    mu0 <- numeric(s$p)
    lambdainv <- matrix(0, s$p, s$p)
  }
  if(!(missing(prior))) {
    tau <- as.numeric(prior[[1]])
    m <- as.numeric(prior[[2]])
    mu0 <- prior[[3]]
    lambdainv <- prior[[4]]
  }
  pri <- numeric(s$d)
  pri[2:(s$p + 1)] <- mu0
  tmp <- as.vector(s$psi[2:(s$p + 1), 2:(s$p + 1)])
  pri[tmp] <- as.vector(lambdainv)
  if(showits)
    cat(paste("Steps of Data Augmentation:", "\n"))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    tmp <- .Fortran("is1n",
		    s$d,
		    start,
		    start,
		    tobs,
		    s$p,
		    s$psi,
		    s$n,
		    s$x,
		    s$npatt,
		    s$r,
		    s$mdpst,
		    s$nmdp,
		    tmpi,
		    tmpi,
		    tmpr,
		    start)
    start <- tmp[[3]]
    start <- .Fortran("ps1n",
		      s$d,
		      start,
		      m,
		      tau,
		      pri,
		      s$p,
		      s$psi,
		      s$n,
		      matrix(0, s$p, s$p),
		      tmpr,
		      start,
		      tmpi)[[5]]
  }
  if(showits)
    cat("\n")
  if(return.ymis) {
    ymis <- tmp[[8]] * matrix(s$sdv, s$n, s$p, T) + matrix(s$xbar, s$n, s$p, T)
    ymis <- ymis[s$ro,  ]
    ymis <- ymis[s$x[s$ro,  ] == 999]
    start <- list(parameter = start, ymis = ymis)
  }
  start
}
"dabipf"<-
function(s, margins, start, steps = 1, prior = 0.5, showits = F)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  w <- !is.na(prior)
  if(any(start[w] == 0)) {
    stop("Starting value on the boundary")
  }
  if(any(!w)) {
    if(any(start[!w] != 0)) {
      stop("Starting value has nonzero elements for structural zeroes")
    }
  }
  start <- as.double(start)
  prior <- as.double(prior)
  prior[!w] <- -999
  if(showits)
    cat(paste("Cycles of Data Augmentation/Bayesian IPF:", "\n"))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    old <- start
    start <- .Fortran("istepc",
		      s$ncells,
		      start,
		      start,
		      s$npatt,
		      s$p,
		      s$r,
		      s$mdpgrp,
		      s$ngrp,
		      s$mobs,
		      s$nmobs,
		      s$d,
		      s$jmp,
		      as.integer(0),
		      s$d,
		      s$d)
    if(start[[13]] == 1)
      stop("Bad parameter value: assigns zero prob. to an observed event")
      else start <- start[[3]]
    start <- .Fortran("bipf",
		      s$ncells,
		      start,
		      old,
		      prior,
		      length(margins),
		      as.integer(margins),
		      s$p,
		      s$d,
		      s$jmp,
		      s$d,
		      s$d,
		      s$d,
		      as.integer(0))
    if(start[[13]] == 1)
      stop("Improper posterior - Check prior.")
    start <- start[[3]]
  }
  if(showits)
    cat("\n")
  start <- start/sum(start)
  array(start, s$d)
}
"dabipf.mix"<-
function(s, margins, design, start, steps = 1, prior = 0.5, showits = F)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  prior <- as.double(prior)
  w <- !is.na(prior)
  prior[!w] <- -999
  storage.mode(design) <- "double"
  storage.mode(margins) <- "integer"
  s$z <- .na.to.snglcode(s$z, 999)
  s$w <- .na.to.icode(s$w, 999)
  tp <- as.integer(s$p)
  tq <- integer(s$q)
  r <- ncol(design)
  npsir <- as.integer((r * (r + 1))/2)
  wk <- numeric(npsir)
  tr <- integer(r)
  wkr <- numeric(r)
  wkd <- numeric(s$ncells)
  beta <- matrix(0, r, s$q)
  psir <- .Fortran("mkpsi",
		   as.integer(r - 1),
		   matrix(as.integer(0), r, r))[[2]]
  tmp <- .Fortran("tobsm",
		  s$q,
		  s$psi,
		  s$npsi,
		  numeric(s$npsi),
		  s$ncells,
		  matrix(0, s$q, s$ncells),
		  numeric(s$ncells),
		  s$npattz,
		  s$rz,
		  s$mdpzgrp,
		  s$npattw,
		  s$p,
		  s$rw,
		  s$mdpwgrp,
		  s$ngrp,
		  s$mobs,
		  s$mobsst,
		  s$nmobs,
		  s$n,
		  s$z,
		  tp,
		  tq)
  kn1 <- tmp[[4]]
  kn2 <- tmp[[6]]
  kn3 <- tmp[[7]]
  sigma <- start$sigma
  pii <- start$pi
  mu <- start$mu
  t1 <- sigma
  t2 <- mu
  t3 <- pii
  if(showits)
    cat(paste("Steps of Data Augmentation-Bayesian IPF:", "\n"))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    tmp <- .Fortran("istepm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    sigma,
		    mu,
		    pii,
		    kn1,
		    kn2,
		    kn3,
		    t1 = t1,
		    t2 = t2,
		    t3 = t3,
		    s$npattz,
		    s$rz,
		    tq,
		    tq,
		    s$mdpzgrp,
		    s$npattw,
		    s$p,
		    s$rw,
		    tp,
		    s$mdpwgrp,
		    s$ngrp,
		    s$mobs,
		    s$mobsst,
		    s$nmobs,
		    s$n,
		    s$z,
		    s$d,
		    s$jmp,
		    tp,
		    kn3,
		    kn1,
		    s$w,
		    numeric(s$q))
    t1 <- tmp$t1
    t2 <- tmp$t2
    t3 <- tmp$t3
    tmp <- .Fortran("pstepcm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    t1,
		    t2,
		    t3,
		    sigma = t1,
		    mu = t2,
		    s$n,
		    r,
		    design,
		    wk,
		    tr,
		    psir,
		    npsir,
		    wkr,
		    wkd,
		    tq,
		    numeric(s$npsi),
		    beta,
		    matrix(0, s$q, s$q))
    sigma <- tmp$sigma
    mu <- tmp$mu
    tmp <- .Fortran("bipf",
		    s$ncells,
		    t3,
		    pii = pii,
		    prior,
		    length(margins),
		    margins,
		    s$p,
		    s$d,
		    s$jmp,
		    tp,
		    tp,
		    tp,
		    err = as.integer(0))
    if(tmp$err == 1)
      stop("Improper posterior - Check prior.")
    pii <- tmp$pii
    pii[w] <- pii[w]/sum(pii[w])
  }
  if(showits)
    cat("\n")
  list(sigma = sigma, mu = mu, pi = pii)
}
"ecm.cat"<-
function(s, margins, start, prior = 1, showits = T, maxits = 1000, eps
	 = 0.0001)
{
  margins <- as.integer(margins)
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  w <- !is.na(prior)
  if(missing(start)) {
    start <- rep(1, s$ncells)
    start[!w] <- 0
  }
    else {
      if(any(start[w] == 0)) {
	warning("Starting value on the boundary")
      }
      if(any(!w)) {
	if(any(start[!w] != 0)) {
	  stop("Starting value has nonzero elements for structural zeroes")
	}
      }
    }
  prior <- as.double(prior)
  start <- as.double(start)
  it <- 0
  convgd <- F
  eps1 <- (1e-007 * sum(s$nmobs))/sum(w)
  if(showits)
    cat(paste("Iterations of ECM:", "\n"))
  d <- integer(s$p)
  while((!convgd) & (it < maxits)) {
    old <- start
    start <- .Fortran("estepc",
		      s$ncells,
		      start,
		      start,
		      s$npatt,
		      s$p,
		      s$r,
		      s$mdpgrp,
		      s$ngrp,
		      s$mobs,
		      s$nmobs,
		      s$d,
		      s$jmp,
		      as.integer(0),
		      integer(s$p),
		      integer(s$p))
    if(start[[13]] == 1)
      stop("Bad parameter value: assigns zero prob. to an observed event"
	   )
      else start <- start[[3]]
    start[w] <- start[w] + prior[w] - 1
    if(any(start[w] < 0))
      stop("Estimate outside the parameter space. Check prior."
	   )
    start <- .Fortran("ipf",
		      s$ncells,
		      as.double(start),
		      as.double(old),
		      length(margins),
		      as.integer(margins),
		      s$p,
		      s$d,
		      s$jmp,
		      d,
		      d,
		      d,
		      eps1)[[3]]
    it <- it + 1
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    convgd <- all(abs(old - start) <= (eps * abs(old)))
  }
  if(showits)
    cat("\n")
  start[w] <- start[w]/sum(start[w])
  array(start, s$d)
}
"ecm.mix"<-
function(s, margins, design, start, prior = 1, maxits = 1000, showits
	 = T, eps = 0.0001)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  prior <- as.double(prior)
  w <- !is.na(prior)
  prior[!w] <- -999
  storage.mode(design) <- "double"
  storage.mode(margins) <- "integer"
  s$z <- .na.to.snglcode(s$z, 999)
  tp <- integer(s$p)
  tq <- integer(s$q)
  sigma <- numeric(s$npsi)
  pii <- numeric(s$ncells)
  mu <- matrix(0, s$q, s$ncells)
  tmp <- .Fortran("tobsm",
		  s$q,
		  s$psi,
		  s$npsi,
		  sigma,
		  s$ncells,
		  mu,
		  pii,
		  s$npattz,
		  s$rz,
		  s$mdpzgrp,
		  s$npattw,
		  s$p,
		  s$rw,
		  s$mdpwgrp,
		  s$ngrp,
		  s$mobs,
		  s$mobsst,
		  s$nmobs,
		  s$n,
		  s$z,
		  tp,
		  tq)
  kn1 <- tmp[[4]]
  kn2 <- tmp[[6]]
  kn3 <- tmp[[7]]
  if(missing(start)) {
    tmp <- .Fortran("stvlm",
		    s$q,
		    s$psi,
		    s$npsi,
		    sigma,
		    s$ncells,
		    mu)
    sigma <- tmp[[4]]
    mu <- tmp[[6]]
    pii <- rep(1, s$ncells)
    pii[!w] <- 0
  }
  if(!missing(start)) {
    sigma <- start$sigma
    mu <- start$mu
    pii <- start$pi
    if(any(pii[w] == 0)) {
      warning("Starting value on the boundary")
    }
    if(any(!w)) {
      if(any(pii[!w] != 0)) {
	stop("Starting value has nonzero elements for structural zeros")
      }
    }
  }
  eps1 <- (1e-007 * s$n)/sum(w)
  r <- ncol(design)
  npsir <- as.integer((r * (r + 1))/2)
  wk <- numeric(npsir)
  tr <- integer(r)
  wkr <- numeric(r)
  wkd <- numeric(s$ncells)
  beta <- matrix(0, r, s$q)
  psir <- .Fortran("mkpsi",
		   as.integer(r - 1),
		   matrix(as.integer(0), r, r))[[2]]
  converged <- F
  it <- 0
  if(showits)
    cat(paste("Steps of ECM:", "\n"))
  t1 <- sigma
  t2 <- mu
  t3 <- pii
  while((!converged) & (it < maxits)) {
    it <- it + 1
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    if(it > 1) {
      sigma <- t1
      mu <- t2
      pii <- t3
    }
    tmp <- .Fortran("estepm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    sigma,
		    mu,
		    pii,
		    kn1,
		    kn2,
		    kn3,
		    t1,
		    t2,
		    t3,
		    s$npattz,
		    s$rz,
		    tq,
		    tq,
		    s$mdpzgrp,
		    s$npattw,
		    s$p,
		    s$rw,
		    tp,
		    s$mdpwgrp,
		    s$ngrp,
		    s$mobs,
		    s$mobsst,
		    s$nmobs,
		    s$n,
		    s$z,
		    s$d,
		    s$jmp,
		    tp,
		    kn3)
    t1 <- tmp[[11]]
    t2 <- tmp[[12]]
    t3 <- tmp[[13]]
    tmp <- .Fortran("mstepcm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    t1,
		    t2,
		    t3,
		    sigma = t1,
		    mu = t2,
		    s$n,
		    r,
		    design,
		    wk,
		    tr,
		    psir,
		    npsir,
		    wkr,
		    wkd,
		    beta)
    t1 <- tmp$sigma
    t2 <- tmp$mu
    t3[w] <- t3[w] + prior[w] - 1
    if(any(t3[w] < 0))
      stop("Estimate outside the parameter space. Check prior."
	   )
    t3 <- .Fortran("ipf",
		   s$ncells,
		   t3,
		   pii,
		   length(margins),
		   margins,
		   s$p,
		   s$d,
		   s$jmp,
		   tp,
		   tp,
		   tp,
		   eps1)[[3]]
    t3[w] <- t3[w]/sum(t3[w])
    c3 <- all(abs(t3 - pii) <= (eps * abs(pii)))
    c2 <- all(abs(t2 - mu) <= (eps * abs(mu)))
    c1 <- all(abs(t1 - sigma) <= (eps * abs(sigma)))
    converged <- c3 & c2 & c1
  }
  if(showits)
    cat("\n")
  list(sigma = t1, mu = t2, pi = t3)
}
"ecme"<-
function(y, subj, occ, pred, xcol, zcol = NULL, vmax, start, maxits = 
	1000, eps = 0.0001, random.effects = F)
{
# Calculates ML estimates for the univariate linear mixed model
# using the ECME algorithm described by Schafer (1997).
  m <- length(table(subj))
  ntot <- length(y)
  nmax <- as.integer(max(occ))
  pcol <- ncol(pred)
  q <- length(zcol)
  p <- length(xcol)			#
  {
    if(missing(vmax)) {
      vmax <- diag(rep(1, nmax))
      iflag <- as.integer(1)
    }
      else iflag <- as.integer(0)
  }
  storage.mode(vmax) <- "double"	#	
  if(is.null(zcol)) {
    q <- as.integer(1)
    tmp <- .Fortran("prelim",
		    ntot,
		    as.integer(subj),
		    m,
		    ist = integer(m),
		    ifin = integer(m),
		    as.integer(occ),
		    nmax,
		    as.double(vmax),
		    vh = array(0, c(nmax, nmax, m)),
		    vi = array(0, c(nmax, nmax, m)),
		    pcol,
		    as.double(pred),
		    q,
		    integer(1),
		    ztv = array(0, c(q, nmax, m)),
		    sig = array(0, c(q, q, m)),
		    iflag)
    ist <- tmp$ist
    ifin <- tmp$ifin
    vh <- tmp$vh
    vi <- tmp$vi
    ztv <- tmp$ztv
    sig0 <- tmp$sig
    tmp <- .Fortran("nopsi",
		    ntot,
		    m,
		    ist,
		    ifin,
		    as.integer(occ),
		    nmax,
		    vi,
		    vh,
		    pcol,
		    as.double(pred),
		    q,
		    integer(1),
		    ztv,
		    sig0,
		    iflag,
		    sig = array(0, c(q, q, m)),
		    psi = matrix(0, q, q),
		    sigma2 = numeric(1),
		    p,
		    as.integer(xcol),
		    beta = numeric(p),
		    matrix(0, q, q),
		    matrix(0, q, q),
		    matrix(0, q, q),
		    as.double(y),
		    delta = numeric(ntot),
		    b = matrix(0, q, m),
		    wk = array(0, c(q, nmax, m)),
		    w = array(0, c(nmax, nmax, m)),
		    xtw = matrix(0, p, nmax),
		    xtwx = matrix(0, p, p),
		    xtwy = numeric(p),
		    xtwxinv = matrix(0, p, p),
		    ll = numeric(1))
    beta <- tmp$beta
    sigma2 <- tmp$sigma2	
#llvec_tmp$ll  I'm not sure that this loglikelihood is right, 
#              so I'll omit it for now.
    llvec <- NULL
    cov.beta <- tmp$xtwxinv
    converged <- T
    iter <- as.integer(1)
    psi <- NULL
    bhat <- NULL
    cov.b <- NULL
  }
    else {
      {
	if(!missing(start)) {
	  beta <- start$beta
	  sigma2 <- start$sigma2
	  psi <- start$psi
	  storage.mode(beta) <- "double"
	  storage.mode(psi) <- "double"
	  storage.mode(sigma2) <- "double"
	  sflag <- as.integer(1)
	}
	  else {
	    beta <- numeric(p)
	    psi <- matrix(0, q, q)
	    sigma2 <- 0
	    sflag <- as.integer(0)
	  }
      }
      llvec <- numeric(as.integer(maxits))
      cat("Performing ECME...")
      tmp <- .Fortran("ecme3",
		      ntot,
		      as.integer(subj),
		      m,
		      ist = integer(m),
		      ifin = integer(m),
		      as.integer(occ),
		      nmax,
		      vi = array(0, c(nmax, nmax, m)),
		      vh = array(0, c(nmax, nmax, m)),
		      pcol,
		      as.double(pred),
		      q,
		      as.integer(zcol),
		      ztv = array(0, c(q, nmax, m)),
		      sig0 = array(0, c(q, q, m)),
		      iflag = iflag,
		      sig = array(0, c(q, q, m)),
		      psi = psi,
		      sigma2 = sigma2,
		      p,
		      as.integer(xcol),
		      beta = beta,
		      matrix(0, q, q),
		      matrix(0, q, q),
		      matrix(0, q, q),
		      as.double(y),
		      delta = rep(0, ntot),
		      b = matrix(0, q, m),
		      wk = array(0, c(q, nmax, m)),
		      w = array(0, c(nmax, nmax, m)),
		      xtw = matrix(0, p, nmax),
		      xtwx = matrix(0, p, p),
		      xtwy = numeric(p),
		      xtwxinv = matrix(0, p, p),
		      llvec = llvec,
		      vmax,
		      sflag = sflag,
		      eps = as.double(eps),
		      obeta = rep(0, p),
		      opsi = matrix(0, q, q),
		      maxits = as.integer(maxits),
		      iter = integer(1),
		      cvgd = integer(1))
      cat("\n")
      iter <- tmp$iter
      llvec <- tmp$llvec[1:iter]
      converged <- tmp$cvgd == as.integer(1)
      cov.beta <- tmp$xtwxinv * tmp$sigma2
      cov.beta <- cov.beta + t(cov.beta) - diag(diag(cov.beta))
      {
	if(random.effects) {
	  bhat <- tmp$b
	  cov.b <- tmp$sig
	  cov.b <- .Fortran("bdiag",
			    q,
			    m,
			    cov.b = cov.b)$cov.b
	}
	  else {
	    bhat <- NULL
	    cov.b <- NULL
	  }
      }
    }
					#
  list(beta = tmp$beta, sigma2 = tmp$sigma2, psi = tmp$psi, 
       converged = converged, iter = iter, loglik = llvec, 
       cov.beta = cov.beta, bhat = bhat, cov.b = cov.b)
}
"em.cat"<-
function(s, start, prior = 1, showits = T, maxits = 1000, eps = 0.0001)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  w <- !is.na(prior)
  if(missing(start)) {
    start <- rep(1, s$ncells)
    start[!w] <- 0
  }
    else {
      if(any(start[w] == 0)) {
	warning("Starting value on the boundary")
      }
      if(any(!w)) {
	if(any(start[!w] != 0)) {
	  stop("Starting value has nonzero elements for structural zeroes")
	}
      }
    }
  prior <- as.double(prior)
  start <- as.double(start)
  it <- 0
  convgd <- F
  if(showits)
    cat(paste("Iterations of EM:", "\n"))
  while((!convgd) & (it < maxits)) {
    old <- start
    start <- .Fortran("estepc",
		      s$ncells,
		      start,
		      start,
		      s$npatt,
		      s$p,
		      s$r,
		      s$mdpgrp,
		      s$ngrp,
		      s$mobs,
		      s$nmobs,
		      s$d,
		      s$jmp,
		      as.integer(0),
		      integer(s$p),
		      integer(s$p))
    if(start[[13]] == 1)
      stop("Bad parameter value: assigns zero prob. to an observed event")
      else start <- start[[3]]
    start[w] <- (start[w] + prior[w] - 1)
    start[w] <- start[w]/sum(start[w])
    if(any(start < 0))
      stop("Estimate outside the parameter space. Check prior.")
    start[w][start[w] < (1e-007/sum(w))] <- 0
    it <- it + 1
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    convgd <- all(abs(old - start) <= (eps * abs(old)))
  }
  if(showits)
    cat("\n")
  array(start, s$d)
}
"em.mix"<-
function(s, start, prior = 1, maxits = 1000, showits = T, eps = 0.0001)
{
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  prior <- as.double(prior)
  w <- !is.na(prior)
  prior[!w] <- -999
  s$z <- .na.to.snglcode(s$z, 999)
  tp <- integer(s$p)
  tq <- integer(s$q)
  sigma <- numeric(s$npsi)
  pii <- numeric(s$ncells)
  mu <- matrix(0, s$q, s$ncells)
  tmp <- .Fortran("tobsm",
		  s$q,
		  s$psi,
		  s$npsi,
		  sigma,
		  s$ncells,
		  mu,
		  pii,
		  s$npattz,
		  s$rz,
		  s$mdpzgrp,
		  s$npattw,
		  s$p,
		  s$rw,
		  s$mdpwgrp,
		  s$ngrp,
		  s$mobs,
		  s$mobsst,
		  s$nmobs,
		  s$n,
		  s$z,
		  tp,
		  tq)
  kn1 <- tmp[[4]]
  kn2 <- tmp[[6]]
  kn3 <- tmp[[7]]
  if(missing(start)) {
    tmp <- .Fortran("stvlm",
		    s$q,
		    s$psi,
		    s$npsi,
		    sigma,
		    s$ncells,
		    mu)
    sigma <- tmp[[4]]
    mu <- tmp[[6]]
    pii <- rep(1, s$ncells)
    pii[!w] <- 0
  }
  if(!missing(start)) {
    sigma <- start$sigma
    mu <- start$mu
    pii <- start$pi
    if(any(pii[w] == 0)) {
      warning("Starting value on the boundary")
    }
    if(any(!w)) {
      if(any(pii[!w] != 0)) {
	stop("Starting value has nonzero elements for structural zeros")
      }
    }
  }
  converged <- F
  it <- 0
  if(showits)
    cat(paste("Steps of EM:", "\n"))
  t1 <- sigma
  t2 <- mu
  t3 <- pii
  while((!converged) & (it < maxits)) {
    it <- it + 1
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    if(it > 1) {
      sigma <- t1
      mu <- t2
      pii <- t3
    }
    tmp <- .Fortran("estepm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    sigma,
		    mu,
		    pii,
		    kn1,
		    kn2,
		    kn3,
		    t1,
		    t2,
		    t3,
		    s$npattz,
		    s$rz,
		    tq,
		    tq,
		    s$mdpzgrp,
		    s$npattw,
		    s$p,
		    s$rw,
		    tp,
		    s$mdpwgrp,
		    s$ngrp,
		    s$mobs,
		    s$mobsst,
		    s$nmobs,
		    s$n,
		    s$z,
		    s$d,
		    s$jmp,
		    tp,
		    kn3)
    t1 <- tmp[[11]]
    t2 <- tmp[[12]]
    t3 <- tmp[[13]]
    tmp <- .Fortran("mstepm",
		    s$q,
		    s$psi,
		    s$npsi,
		    s$ncells,
		    t1,
		    t2,
		    t3,
		    s$n,
		    prior)
    t1 <- tmp[[5]]
    t2 <- tmp[[6]]
    t3 <- tmp[[7]]
    if(any(t3 < 0))
      stop("Estimate outside the parameter space. Check prior.")
    t3[w][t3[w] < (1e-007/sum(w))] <- 0
    c3 <- all(abs(t3 - pii) <= (eps * abs(pii)))
    c2 <- all(abs(t2 - mu) <= (eps * abs(mu)))
    c1 <- all(abs(t1 - sigma) <= (eps * abs(sigma)))
    converged <- c3 & c2 & c1
  }
  if(showits)
    cat("\n")
  list(sigma = t1, mu = t2, pi = t3)
}
"em.norm"<-
function(s, start, showits = T, maxits = 1000, criterion = 0.0001, 
	prior)
{
  s$x <- .na.to.snglcode(s$x, 999)
  if(missing(start)) {
    start <- .Fortran("stvaln",
		      s$d,
		      numeric(s$d),
		      s$p,
		      s$psi)[[2]]
  }
  if(missing(prior)) {
    mle <- as.integer(1)
    tau <- numeric(1)
    m <- numeric(1)
    mu0 <- numeric(s$p)
    lambdainv <- matrix(0, s$p, s$p)
  }
  if(!(missing(prior))) {
    mle <- as.integer(0)
    tau <- as.numeric(prior[[1]])
    m <- as.numeric(prior[[2]])
    mu0 <- as.numeric(prior[[3]])
    lambdainv <- as.numeric(prior[[4]])
  }
  tmp <- as.integer(numeric(s$p))
  tobs <- .Fortran("tobsn",
		   s$d,
		   numeric(s$d),
		   s$p,
		   s$psi,
		   s$n,
		   s$x,
		   s$npatt,
		   s$r,
		   s$mdpst,
		   s$nmdp,
		   tmp)[[2]]		# iterate to mle
  it <- 0
  converged <- F
  if(showits)
    cat(paste("Iterations of EM:", "\n"))
  while((!converged) & (it < maxits)) {
    old <- start
    start <- .Fortran("emn",
		      s$d,
		      old,
		      start,
		      tobs,
		      s$p,
		      s$psi,
		      s$n,
		      s$x,
		      s$npatt,
		      s$r,
		      s$mdpst,
		      s$nmdp,
		      tmp,
		      tmp,
		      numeric(s$p),
		      mle,
		      tau,
		      m,
		      mu0,
		      lambdainv)[[3]]	# print iteration number
    it <- it + 1
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    converged <- max(abs(old - start)) <= criterion
  }
  if(showits)
    cat("\n")
  start
}
"getparam.mix"<-
function(s, theta, corr = F)
{
  pii <- array(theta$pi, s$d)
  if(!is.null(s$wnames)) {
    pinames <- as.list(1:s$p)
    for(i in 1:s$p) {
      pinames[[i]] <- paste(s$wnames[i], "=", format(1:s$d[i]), sep = "")
    }
    dimnames(pii) <- pinames
  }
  mu <- theta$mu * s$sdv + s$xbar
  dimnames(mu) <- list(s$znames, NULL)
  sigma <- theta$sigma[s$psi]
  sigma <- matrix(sigma, s$q, s$q)
  tmp <- matrix(s$sdv, s$q, s$q)
  sigma <- sigma * tmp * t(tmp)
  dimnames(sigma) <- list(s$znames, s$znames)
  mu[, pii == 0] <- NA
  if(corr) {
    sdv <- sqrt(diag(sigma))
    names(sdv) <- s$znames
    tmp <- matrix(sdv, s$q, s$q)
    r <- sigma/(tmp * t(tmp))
    dimnames(r) <- list(s$znames, s$znames)
    result <- list(pi = pii, mu = mu, sdv = sdv, r = r)
  }
    else result <- list(pi = pii, mu = mu, sigma = sigma)
  result
}
"getparam.norm"<-
function(s, theta, corr = F)
{
  mu <- theta[s$psi[1, 2:(s$p + 1)]] * s$sdv + s$xbar
  names(mu) <- dimnames(s$x)[[2]]
  sigma <- theta[s$psi[2:(s$p + 1), 2:(s$p + 1)]]
  sigma <- matrix(sigma, s$p, s$p)
  tmp <- matrix(s$sdv, s$p, s$p)
  sigma <- sigma * tmp * t(tmp)
  dimnames(sigma) <- list(names(mu), names(mu))
  if(corr) {
    sdv <- sqrt(diag(sigma))
    names(sdv) <- names(mu)
    tmp <- matrix(sdv, s$p, s$p)
    r <- sigma/(tmp * t(tmp))
    dimnames(r) <- list(names(mu), names(mu))
    result <- list(mu = mu, sdv = sdv, r = r)
  }
    else result <- list(mu = mu, sigma = sigma)
  result
}
"imp.cat"<-
function(s, theta)
{
  theta <- as.double(theta/sum(theta))
  if(!(s$grouped)) {
    x <- s$x
    x[is.na(x)] <- as.integer(0)
    x <- .Fortran("impc",
		  s$n,
		  s$p,
		  x,
		  s$ncells,
		  theta,
		  s$npatt,
		  s$r,
		  s$mdpgst,
		  s$mdpgrp,
		  s$ngrp,
		  s$mobs,
		  s$nmobs,
		  s$mobsst,
		  s$d,
		  s$jmp,
		  s$d,
		  s$d)[[3]]
    result <- x[s$ro,  ]
  }
  if(s$grouped) {
    tab <- .Fortran("istepc",
		    s$ncells,
		    theta,
		    theta,
		    s$npatt,
		    s$p,
		    s$r,
		    s$mdpgrp,
		    s$ngrp,
		    s$mobs,
		    s$nmobs,
		    s$d,
		    s$jmp,
		    as.integer(0),
		    s$d,
		    s$d)[[3]]
    tab <- as.integer(round(tab))
    tmp <- array(integer(1), s$d)
    levmat <- matrix(integer(1), s$ncells, s$p)
    for(j in 1:s$p)
      levmat[, j] <- slice.index(tmp, j)
    result <- list(x = levmat[tab != 0,  ], counts = tab[
					      tab != 0])
  }
  result
}
"imp.mix"<-
function(s, theta, x)
{
	sigma <- theta$sigma
	mu <- theta$mu
	pii <- theta$pi
	z <- .na.to.snglcode(s$z, 999)
	w <- .na.to.icode(s$w, 999)
	tp <- as.integer(numeric(s$p))
	tq <- as.integer(numeric(s$q))
	tmp <- .Fortran("istepm",
		s$q,
		s$psi,
		s$npsi,
		s$ncells,
		sigma,
		mu,
		pii,
		sigma,
		mu,
		pii,
		sigma,
		mu,
		pii,
		s$npattz,
		s$rz,
		tq,
		tq,
		s$mdpzgrp,
		s$npattw,
		s$p,
		s$rw,
		tp,
		s$mdpwgrp,
		s$ngrp,
		s$mobs,
		s$mobsst,
		s$nmobs,
		s$n,
		z = z,
		s$d,
		s$jmp,
		tp,
		pii,
		sigma,
		w = w,
		numeric(s$q))
	w <- tmp$w[s$ro,  ]
	z <- tmp$z * matrix(s$sdv, s$n, s$q, T) + matrix(s$xbar, s$n, 
		s$q, T)
	z <- z[s$ro,  ]
	if(!missing(x)) {
		zorig <- x[, (s$p + 1):(s$p + s$q)]
		z[!is.na(zorig)] <- zorig[!is.na(zorig)]
	}
	ximp <- cbind(w, z)
	dimnames(ximp) <- list(s$rnames, c(s$wnames, s$znames))
	ximp
}
"imp.norm"<-
function(s, theta, x)
{
  s$x <- .na.to.snglcode(s$x, 999)
  tmpi <- as.integer(numeric(s$p))
  tmpr <- as.single((numeric(s$p)))
  tobs <- .Fortran("tobsn",
		   s$d,
		   numeric(s$d),
		   s$p,
		   s$psi,
		   s$n,
		   s$x,
		   s$npatt,
		   s$r,
		   s$mdpst,
		   s$nmdp,
		   tmpi)[[2]]
  s$x <- .Fortran("is1n",
		  s$d,
		  theta,
		  theta,
		  tobs,
		  s$p,
		  s$psi,
		  s$n,
		  s$x,
		  s$npatt,
		  s$r,
		  s$mdpst,
		  s$nmdp,
		  tmpi,
		  tmpi,
		  tmpr,
		  theta)[[8]]
  s$x <- s$x * matrix(s$sdv, s$n, s$p, T) + matrix(s$xbar, s$n, s$p, T)
  s$x <- s$x[s$ro,  ]
  if(!missing(x))
    x[is.na(x)] <- s$x[is.na(x)]
    else {
      x <- s$x
      storage.mode(x) <- "double"
    }
  x
}
"ipf"<-
function(table, margins, start, eps = 0.0001, maxits = 50, showits = T
	)
{
  a <- attributes(table)
  d <- dim(table)
  n <- length(table)
  j <- as.integer(cumprod(d)/d)
  if(!missing(start)) {
    if(length(start) != n)
      stop("Starting value has incorrect length")
  }
  if(missing(start))
    start <- rep(1, n)
  convgd <- F
  it <- 0
  if(showits)
    cat(paste("Cycles of IPF:", "\n", sep = ""))
  eps1 <- 1e-007 * mean(table)
  while((!convgd) & (it < maxits)) {
    old <- start
    it <- it + 1
    if(showits)
      cat(paste(format(it), "...", sep = ""))
    start <- .Fortran("ipf",
		      n,
		      as.double(table),
		      as.double(old),
		      length(margins),
		      as.integer(margins),
		      length(d),
		      d,
		      j,
		      d,
		      d,
		      d,
		      eps1)[[3]]
    convgd <- all(abs(old - start) <= (eps * abs(old)))
  }
  if(showits)
    cat("\n")
  attributes(start) <- a
  start
}
"loglik.mix"<-
function(s, theta)
{
  s$z <- .na.to.snglcode(s$z, 999)
  tp <- as.integer(numeric(s$p))
  tq <- as.integer(numeric(s$q))
  ll <- .Fortran("lobsm",
		 s$q,
		 s$psi,
		 s$npsi,
		 s$ncells,
		 theta$sigma,
		 theta$mu,
		 theta$pi,
		 s$npattz,
		 s$rz,
		 tq,
		 tq,
		 s$mdpzgrp,
		 s$npattw,
		 s$p,
		 s$rw,
		 tp,
		 s$mdpwgrp,
		 s$ngrp,
		 s$mobs,
		 s$mobsst,
		 s$nmobs,
		 s$n,
		 s$z,
		 s$d,
		 s$jmp,
		 integer(s$p),
		 ll = numeric(1))$ll
  ll
}
"loglik.norm"<-
function(s, theta)
{
  s$x <- .na.to.snglcode(s$x, 999)
  .Fortran("lobsn",
	   s$d,
	   theta,
	   numeric(s$d),
	   s$p,
	   s$psi,
	   s$n,
	   s$x,
	   s$npatt,
	   s$r,
	   s$mdpst,
	   s$nmdp,
	   as.integer(numeric(s$p)),
	   numeric(s$p),
	   0)[[14]]
}
"logpost.cat"<-
function(s, theta, prior)
{
  tmp <- .Fortran("llc",
		  s$ncells,
		  as.double(theta),
		  s$npatt,
		  s$p,
		  s$r,
		  s$mdpgrp,
		  s$ngrp,
		  s$mobs,
		  s$nmobs,
		  s$d,
		  s$jmp,
		  s$d,
		  s$d,
		  numeric(1),
		  as.integer(0))
  if(tmp[[15]] == 1) {
    stop("Bad parameter: has zero probability for a nonzero cell"
	 )
  }
  lp <- tmp[[14]]
  if(!missing(prior)) {
    lp <- lp + sum((prior[!is.na(prior)] - 1) * log(theta[!is.na(prior)]))
  }
  lp
}
"logpost.norm"<-
function(s, theta, prior)
{
  s$x <- .na.to.snglcode(s$x, 999)
  l1 <- .Fortran("lobsn",
		 s$d,
		 theta,
		 numeric(s$d),
		 s$p,
		 s$psi,
		 s$n,
		 s$x,
		 s$npatt,
		 s$r,
		 s$mdpst,
		 s$nmdp,
		 as.integer(numeric(s$p)),
		 numeric(s$p),
		 0)[[14]]
  if(!(missing(prior))) {
    tau <- as.numeric(prior[[1]])
    m <- as.numeric(prior[[2]])
    mu0 <- as.numeric(prior[[3]])
    lambdainv <- as.numeric(prior[[4]])
  }
  if(missing(prior)) {
    tau <- as.numeric(0)
    m <- as.numeric(-1)
    mu0 <- numeric(s$p)
    lambdainv <- matrix(0, s$p, s$p)
  }
  l2 <- .Fortran("lprin",
		 s$d,
		 theta,
		 s$p,
		 s$psi,
		 numeric(s$p),
		 tau,
		 m,
		 mu0,
		 lambdainv,
		 0)[[10]]
  l1 + l2
}
"makeparam.norm"<-
function(s, thetalist)
{
  result <- numeric(s$d)
  result[1] <- -1
  xbar <- s$xbar
  sdv <- s$sdv
  mu <- (thetalist[[1]] - xbar)/sdv
  result[2:(s$p + 1)] <- mu
  if(length(thetalist) == 3) {
    tmp <- matrix(thetalist[[2]], s$p, s$p)
    sigma <- thetalist[[3]] * tmp * t(tmp)
  }
    else sigma <- thetalist[[2]]
  tmp <- matrix(sdv, s$p, s$p)
  sigma <- sigma/(tmp * t(tmp))
  tmp <- as.vector(s$psi[2:(s$p + 1), 2:(s$p + 1)])
  result[tmp] <- as.vector(sigma)
  result
}
"mda.cat"<-
function(s, start, steps = 1, prior = 0.5, showits = F)
{
  start <- as.double(start/sum(start))
  if(length(prior) == 1)
    prior <- rep(prior, s$ncells)
  if((!missing(prior)) & (length(prior) == 1)) {
    prior <- as.double(rep(prior, s$ncells))
  }
  if((!missing(prior)) & any(is.na(prior))) {
    if(any(start[is.na(prior)] != 0)) {
      stop("Bad starting value: nonzero elements for structural zeroes")
    }
  }
  if(showits)
    cat(paste("Steps of Monotone Data Augmentation:", "\n"))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    start <- .Fortran("mgstepc",
		      s$ncells,
		      start,
		      start,
		      s$npatt,
		      s$p,
		      s$r,
		      s$mdpgrp,
		      s$ngrp,
		      s$mobs,
		      s$nmobs,
		      s$d,
		      s$jmp,
		      s$d,
		      s$d,
		      prior,
		      prior,
		      as.integer(0),
		      s$sj)
    if(start[[17]] == 1)
      stop("Bad parameter value: assigns zero prob. to an observed event"
	   )
    if(start[[17]] == 2)
      stop("Improper posterior distribution")
    start <- start[[3]]
    start <- start/sum(start)
  }
  if(showits)
    cat("\n")
  array(start, s$d)
}
"mda.norm"<-
function(s, start, steps = 1, showits = F)
{
  s$x <- .na.to.snglcode(s$x, 999)
  tobs <- .Fortran("tobsmn",
		   s$p,
		   s$psi,
		   s$n,
		   s$x,
		   s$npatt,
		   s$r,
		   s$mdpst,
		   s$nmdp,
		   s$last,
		   integer(s$p),
		   s$sj,
		   s$layer,
		   s$nlayer,
		   s$d,
		   matrix(0, s$nlayer, s$d))[[15]]
  if(showits)
    cat(paste("Steps of Monotone Data Augmentation:", "\n"
	      ))
  for(i in 1:steps) {
    if(showits)
      cat(paste(format(i), "...", sep = ""))
    s$x <- .Fortran("is2n",
		    s$d,
		    theta,
		    s$p,
		    s$psi,
		    s$n,
		    s$x,
		    s$npatt,
		    s$r,
		    s$mdpst,
		    s$nmdp,
		    s$sj,
		    s$last,
		    integer(s$p),
		    integer(s$p),
		    single(s$p),
		    theta)[[6]]
    theta <- .Fortran("ps2n",
		      s$p,
		      s$psi,
		      s$n,
		      s$x,
		      s$npatt,
		      s$r,
		      s$mdpst,
		      s$nmdp,
		      integer(s$p),
		      integer(s$p),
		      s$nmon,
		      s$sj,
		      s$nlayer,
		      s$d,
		      tobs,
		      numeric(s$d),
		      numeric(s$d),
		      numeric(s$p + 1),
		      numeric(s$d))[[19]]
  }
  if(showits)
    cat("\n")
  theta
}
"mi.inference"<-
function(est, std.err, confidence = 0.95)
{
  qstar <- est[[1]]
  for(i in 2:length(est)) {
    qstar <- cbind(qstar, est[[i]])
  }
  qbar <- apply(qstar, 1, mean)
  u <- std.err[[1]]
  for(i in 2:length(std.err)) {
    u <- cbind(u, std.err[[i]])
  }
  dimnames(u)[[1]] <- dimnames(qstar)[[1]]
  u <- u^2
  ubar <- apply(u, 1, mean)
  bm <- apply(qstar, 1, var)
  m <- dim(qstar)[2]
  tm <- ubar + ((1 + (1/m)) * bm)
  rem <- ((1 + (1/m)) * bm)/ubar
  nu <- (m - 1) * (1 + (1/rem))^2
  alpha <- 1 - (1 - confidence)/2
  low <- qbar - qt(alpha, nu) * sqrt(tm)
  up <- qbar + qt(alpha, nu) * sqrt(tm)
  pval <- 2 * (1 - pt(abs(qbar/sqrt(tm)), nu))
  fminf <- (rem + 2/(nu + 3))/(rem + 1)
  result <- list(est = qbar, std.err = sqrt(tm), df = nu, signif
		 = pval, lower = low, upper = up, r = rem, fminf = 
		 fminf)
  result
}
"ninvwish"<-
function(s, params)
{
  tau <- as.numeric(params[[1]])
  m <- as.numeric(params[[2]])
  mu0 <- params[[3]]
  lambdainv <- params[[4]]
  tmpi <- as.integer(numeric(s$p))
  tmpr <- as.single((numeric(s$p)))
  pri <- numeric(s$d)
  pri[2:(s$p + 1)] <- mu0
  tmp <- as.vector(s$psi[2:(s$p + 1), 2:(s$p + 1)])
  pri[tmp] <- as.vector(lambdainv)
  .Fortran("ninvwn",
	   s$d,
	   pri,
	   tau,
	   m,
	   s$p,
	   s$psi,
	   numeric((s$p)^2),
	   tmpr,
	   numeric(s$d),
	   tmpi)[[2]]
}
"pan"<-
function(y, subj, pred, xcol, zcol, prior, seed, iter = 1, start)
{
  if(any(is.na(pred))) stop("missing values in pred not allowed")	
# change y and pred to matrices, if necessary
  if(is.vector(y))
    y <- matrix(y, ncol = 1)
  if(is.vector(pred)) pred <- matrix(pred, ncol = 1)	
# change prior$Binv and Dinv to matrices if necessary
  if(length(prior$Binv) == 1)
    prior$Binv <- matrix(prior$Binv, 1, 1)
  if(length(prior$Dinv) == 1)
    prior$Dinv <- matrix(prior$Binv, 1, 1)
  m <- as.integer(length(table(subj)))
  ntot <- as.integer(nrow(y))
  r <- as.integer(ncol(y))
  p <- length(xcol)
  q <- length(zcol)
  zcol <- as.integer(zcol)
  xcol <- as.integer(xcol)
  pcol <- as.integer(ncol(pred))	#
  {
    if(missing(start)) {
      beta <- matrix(0, p, r)
      sigma <- matrix(0, r, r)
      psi <- matrix(0, q * r, q * r)
      eps <- matrix(0, ntot, r)
      sflag <- as.integer(0)
    }
      else {
	beta <- start$beta
	sigma <- start$sigma
	psi <- start$psi
	eps <- start$y
	sflag <- as.integer(1)
	storage.mode(eps) <- "double"
	storage.mode(beta) <- "double"
	storage.mode(sigma) <- "double"
	storage.mode(psi) <- "double"
      }
  }
#
  nhyp <- as.integer(2 + ((r * (r + 1))/2) + ((r * q * (r * q + 1))/2))
  hyp <- rep(0, nhyp)
  hyp[1] <- prior$a
  k <- 1
  for(i in 1:r) {
    for(j in i:r) {
      k <- k + 1
      hyp[k] <- prior$Binv[i, j]
    }
  }
  k <- k + 1
  hyp[k] <- prior$c
  for(i in 1:(r * q)) {
    for(j in i:(r * q)) {
      k <- k + 1
      hyp[k] <- prior$Dinv[i, j]
    }
  }
  storage.mode(hyp) <- "double"	
# create rmat, npatt and patt to keep track of missingness patterns
  rmat <- 1 - 1 * is.na(y)
  storage.mode(rmat) <- "integer"
  revcpatt <- rep("", ntot)
  for(i in 1:r)
    revcpatt <- paste(as.character(rmat[, i]), revcpatt, sep = "")
  nulpat0 <- ""
  nulpat2 <- ""
  for(i in 1:r) {
    nulpat0 <- paste(nulpat0, "0", sep = "")
    nulpat2 <- paste(nulpat2, "2", sep = "")
  }
  revcpatt[revcpatt == nulpat0] <- nulpat2
  tmp <- rev(table(revcpatt))
  npatt <- length(tmp)
  if(any(revcpatt == nulpat2))
    npatt <- npatt - 1
  w <- !duplicated(revcpatt)
  upatt <- revcpatt[w]
  rmat <- rmat[w,  ]
  if(r == 1)
    rmat <- matrix(rmat, ncol = 1)
  w <- rev(order(upatt))
  upatt <- upatt[w]
  rmat <- rmat[w,  ]
  if(r == 1)
    rmat <- matrix(rmat, ncol = 1)
  if(any(upatt == nulpat2)) {
    rmat <- rmat[-1,  ]
    upatt <- upatt[-1]
  }
  patt <- integer(ntot)
  patt[revcpatt == nulpat2] <- 0
  for(i in 1:npatt)
    patt[revcpatt == upatt[i]] <- i
  storage.mode(npatt) <- "integer"
  storage.mode(rmat) <- "integer"
  storage.mode(patt) <- "integer"
  iposn <- as.integer(1:ntot)
  w <- order(patt)
  iposn <- iposn[w]
  pstfin <- matrix(0, npatt, 2)
  {
    if(any(patt == 0)) {
      st <- tmp[1] + 1
      for(i in 2:(npatt + 1)) {
	pstfin[i - 1, 1] <- st
	pstfin[i - 1, 2] <- st + tmp[i] - 1
	st <- st + tmp[i]
      }
    }
      else {
	st <- 1
	for(i in 1:npatt) {
	  pstfin[i, 1] <- st
	  pstfin[i, 2] <- st + tmp[i] - 1
	  st <- st + tmp[i]
	}
      }
  }
  storage.mode(pstfin) <- "integer"	#
  storage.mode(y) <- "double"
  y[is.na(y)] <- -999.99
  storage.mode(pred) <- "double"	#
  .Fortran("rngs", as.integer(seed))		#
  tmp <- .Fortran("mgibbs",
		  ntot,
		  as.integer(subj),
		  m,
		  ist = integer(m),
		  ifin = integer(m),
		  pcol,
		  pred,
		  q,
		  zcol,
		  ztz = array(0, c(q, q, m)),
		  patt,
		  nstar = integer(1),
		  r,
		  y = y,
		  p,
		  xcol,
		  npatt,
		  rmat,
		  sflag,
		  beta = beta,
		  sigma = sigma,
		  psi = psi,
		  b = array(0, c(q, r, m)),
		  xtxinv = matrix(0, p, p),
		  wkpp = matrix(0, p, p),
		  wkpr = matrix(0, p, r),
		  eps = eps,
		  wkrqrq1 = matrix(0, r * q, r * q),
		  wkrqrq2 = matrix(0, r * q, r * q),
		  sig = array(0, c(r * q, r * q, m)),
		  wkrr1 = matrix(0, r, r),
		  wkrr2 = matrix(0, r, r),
		  iter = as.integer(iter),
		  wkqr1 = matrix(0, q, r),
		  wkqr2 = matrix(0, q, r),
		  wkqrv = numeric(q * r),
		  nhyp,
		  hyp,
		  delta = matrix(0, ntot, r),
		  iposn = iposn,
		  pstfin = pstfin,
		  betas = array(0, c(p, r, iter)),
		  sigmas = array(0, c(r, r, iter)),
		  psis = array(0, c(r * q, r * q, iter)))	
# restore y to being a vector if necessary
  if(r == 1) tmp$y <- as.vector(tmp$y)	#
  list(beta = tmp$betas, sigma = tmp$sigmas, psi = tmp$psis, y
       = tmp$y, last = list(beta = tmp$beta, sigma = tmp$
		  sigma, psi = tmp$psi, y = tmp$y))
}
"pan.bd"<-
function(y, subj, pred, xcol, zcol, prior, seed, iter = 1, start)
{
  if(any(is.na(pred))) stop("missing values in pred not allowed")	
# change y and pred to matrices, if necessary
  if(is.vector(y))
    y <- matrix(y, ncol = 1)
  if(is.vector(pred)) pred <- matrix(pred, ncol = 1)	
# change prior$Binv and Dinv to matrices if necessary
  if(length(prior$Binv) == 1)
    prior$Binv <- matrix(prior$Binv, 1, 1)
  if(length(prior$Dinv) == 1)
    prior$Dinv <- matrix(prior$Binv, 1, 1)
  m <- as.integer(length(table(subj)))
  ntot <- as.integer(nrow(y))
  r <- as.integer(ncol(y))
  p <- length(xcol)
  q <- length(zcol)
  zcol <- as.integer(zcol)
  xcol <- as.integer(xcol)
  pcol <- as.integer(ncol(pred))	#
  {
    if(missing(start)) {
      beta <- matrix(0, p, r)
      sigma <- matrix(0, r, r)
      psi <- array(0, c(q, q, r))
      eps <- matrix(0, ntot, r)
      sflag <- as.integer(0)
    }
      else {
	beta <- start$beta
	sigma <- start$sigma
	psi <- start$psi
	eps <- start$y
	sflag <- as.integer(1)
	storage.mode(eps) <- "double"
	storage.mode(beta) <- "double"
	storage.mode(sigma) <- "double"
	storage.mode(psi) <- "double"
      }
  }
# This part packs the hyperparameters a, Binv, c, Dinv into a vector
# of length 1 + r*(r+1)/2 + r + r*q*(q+1)/2
  nhyp <- as.integer(1 + (r * (r + 1))/2 + r + (r * q * (q + 1))/ 2)
  hyp <- rep(0, nhyp)
  hyp[1] <- prior$a
  k <- 1
  for(i in 1:r) {
    for(j in i:r) {
      k <- k + 1
      hyp[k] <- prior$Binv[i, j]
    }
  }
  for(i in 1:r) {
    k <- k + 1
    hyp[k] <- prior$c[i]
  }
  for(l in 1:r) {
    for(i in 1:q) {
      for(j in i:q) {
	k <- k + 1
	hyp[k] <- prior$Dinv[i, j, l]
      }
    }
  }
  storage.mode(hyp) <- "double"	
# create rmat, npatt and patt to keep track of missingness patterns
  rmat <- 1 - 1 * is.na(y)
  storage.mode(rmat) <- "integer"
  revcpatt <- rep("", ntot)
  for(i in 1:r)
    revcpatt <- paste(as.character(rmat[, i]), revcpatt, sep = "")
  nulpat0 <- ""
  nulpat2 <- ""
  for(i in 1:r) {
    nulpat0 <- paste(nulpat0, "0", sep = "")
    nulpat2 <- paste(nulpat2, "2", sep = "")
  }
  revcpatt[revcpatt == nulpat0] <- nulpat2
  tmp <- rev(table(revcpatt))
  npatt <- length(tmp)
  if(any(revcpatt == nulpat2))
    npatt <- npatt - 1
  w <- !duplicated(revcpatt)
  upatt <- revcpatt[w]
  rmat <- rmat[w,  ]
  if(r == 1)
    rmat <- matrix(rmat, ncol = 1)
  w <- rev(order(upatt))
  upatt <- upatt[w]
  rmat <- rmat[w,  ]
  if(r == 1)
    rmat <- matrix(rmat, ncol = 1)
  if(any(upatt == nulpat2)) {
    rmat <- rmat[-1,  ]
    upatt <- upatt[-1]
  }
  patt <- integer(ntot)
  patt[revcpatt == nulpat2] <- 0
  for(i in 1:npatt)
    patt[revcpatt == upatt[i]] <- i
  storage.mode(npatt) <- "integer"
  storage.mode(rmat) <- "integer"
  storage.mode(patt) <- "integer"
  iposn <- as.integer(1:ntot)
  w <- order(patt)
  iposn <- iposn[w]
  pstfin <- matrix(0, npatt, 2)
  {
    if(any(patt == 0)) {
      st <- tmp[1] + 1
      for(i in 2:(npatt + 1)) {
	pstfin[i - 1, 1] <- st
	pstfin[i - 1, 2] <- st + tmp[i] - 1
	st <- st + tmp[i]
      }
    }
      else {
	st <- 1
	for(i in 1:npatt) {
	  pstfin[i, 1] <- st
	  pstfin[i, 2] <- st + tmp[i] - 1
	  st <- st + tmp[i]
	}
      }
  }
  storage.mode(pstfin) <- "integer"	#
  storage.mode(y) <- "double"
  y[is.na(y)] <- -999.99
  storage.mode(pred) <- "double"	#
  .Fortran("rngs",
	   as.integer(seed))		#
  tmp <- .Fortran("mgibbsbd",
		  ntot,
		  as.integer(subj),
		  m,
		  ist = integer(m),
		  ifin = integer(m),
		  pcol,
		  pred,
		  q,
		  zcol,
		  ztz = array(0, c(q, q, m)),
		  patt,
		  nstar = integer(1),
		  r,
		  y = y,
		  p,
		  xcol,
		  npatt,
		  rmat,
		  sflag,
		  beta = beta,
		  sigma = sigma,
		  psi = psi,
		  b = array(0, c(q, r, m)),
		  xtxinv = matrix(0, p, p),
		  wkpp = matrix(0, p, p),
		  wkpr = matrix(0, p, r),
		  eps = eps,
		  wkrqrq1 = matrix(0, r * q, r * q),
		  wkrqrq2 = matrix(0, r * q, r * q),
		  sig = array(0, c(r * q, r * q, m)),
		  wkrr1 = matrix(0, r, r),
		  wkrr2 = matrix(0, r, r),
		  iter = as.integer(iter),
		  wkqr1 = matrix(0, q, r),
		  wkqr2 = matrix(0, q, r),
		  wkqrv = numeric(q * r),
		  nhyp = nhyp,
		  hyp = hyp,
		  delta = matrix(0, ntot, r),
		  iposn = iposn,
		  pstfin = pstfin,
		  betas = array(0, c(p, r, iter)),
		  sigmas = array(0, c(r, r, iter)),
		  psis = array(0, c(q, q, r, iter)),
		  wkqq1 = matrix(0, q, q),
		  wkqq2 = matrix(0, q, q))	
# restore y to being a vector if necessary
  if(r == 1) tmp$y <- as.vector(tmp$y)	#
  list(beta = tmp$betas, sigma = tmp$sigmas, psi = tmp$psis, y
       = tmp$y, last = list(beta = tmp$beta, sigma = tmp$
		  sigma, psi = tmp$psi, y = tmp$y))
}
"prelim.cat"<-
function(x, counts, levs)
{
# get dimensions of x
  if(is.vector(x)) x <- matrix(x, length(x), 1)
  p <- ncol(x)
  n <- nrow(x)
  if(missing(counts)) {
    counts <- rep(1, nrow(x))
    grouped <- F
  }
  if(!missing(counts)) {
    grouped <- T
  }
# get dimensions of contingency table
  if(missing(levs)) {
    d <- as.integer(numeric(p))
    for(j in 1:p) {
      d[j] <- as.integer(max(x[!is.na(x[, j]), j]))
    }
  }
  if(!(missing(levs))) {
    d <- as.integer(levs)
  }
# get response indicators
  r <- 1 * is.na(x)
  nmis <- as.integer(apply(r, 2, sum))
  names(nmis) <- dimnames(x)[[2]]	
# index the missing data patterns
  mdp <- as.integer((r %*% (2^((1:ncol(x)) - 1))) + 1)	
# calculate the known part of the cell number for rows of x
  cumd <- as.integer(cumprod(d))
  ncells <- cumd[p]
  jmp <- as.integer(cumd/d)
  r <- 1 - r
  tmp <- x
  tmp[is.na(tmp)] <- as.integer(1)
  mobs0 <- as.integer(1 + ((tmp - 1) * r) %*% (cumd/d))	
# do row sort
  ro <- order(mdp, mobs0)
  x <- as.matrix(x[ro,  ], n, p)
  mdp <- mdp[ro]
  mobs0 <- mobs0[ro]
  r <- as.matrix(r[ro,  ], n, p)
  counts <- counts[ro]
  ro <- order(ro)
  storage.mode(x) <- "integer"		# compress missing data patterns
  mdpst <- as.integer(seq(along = mdp)[!duplicated(mdp)])
  mdp <- unique(mdp)
  npatt <- length(mdp)
  if(npatt > 1)
    nmdp <- as.integer(c(mdpst[-1], n + 1) - mdpst)
  if(npatt == 1) nmdp <- n	
# group the data within missing data patterns
  mdpgrp <- numeric(npatt)
  for(j in 1:npatt) {
    mdpgrp[j] <- length(table(mobs0[mdpst[j]:(mdpst[j] + 
					      nmdp[j] - 1)]))
  }
  mdpgrp <- as.integer(mdpgrp)
  ngrp <- as.integer(sum(mdpgrp))
  if(ngrp == 1)
    mdpgst <- as.integer(1)
  if(ngrp > 1)
    mdpgst <- as.integer(cumsum(c(1, mdpgrp[ - npatt])))
  mobs <- numeric(ngrp)
  mobsst <- numeric(ngrp)
  nmobs <- numeric(ngrp)
  for(i in 1:npatt) {
    w <- mdpst[i]:(mdpst[i] + nmdp[i] - 1)
    g <- mdpgst[i]:(mdpgst[i] + mdpgrp[i] - 1)
    mobs[g] <- unique(mobs0[w])
    mobsst[g] <- w[!duplicated(mobs0[w])]
    for(j in g) {
      nmobs[j] <- sum((counts[w])[mobs0[w] == mobs[j
					 ]])
    }
  }
  storage.mode(mobs) <- "integer"
  storage.mode(mobsst) <- "integer"
  storage.mode(nmobs) <- "integer"	
# create r-matrix for display purposes
  r <- matrix(r[mdpst,  ], npatt, p)
  if(!grouped) {
    if(npatt == 1)
      tmp <- format(n)
    if(npatt > 1)
      tmp <- format(c(mdpst[2:npatt], n + 1) - mdpst)
  }
  if(grouped) {
    tmp <- numeric(npatt)
    for(i in 1:npatt) {
      w <- mdpgst[i]:(mdpgst[i] + mdpgrp[i] - 1)
      tmp[i] <- sum(nmobs[w])
      tmp <- format(tmp)
    }
  }
  dimnames(r) <- list(tmp, dimnames(x)[[2]])
  storage.mode(r) <- "integer"	
# find s_j's for monotone algorithm
  sj <- .Fortran("sjn",
		 p,
		 npatt,
		 r,
		 integer(p))[[4]]	# return list
  list(x = x, n = n, p = p, d = d, jmp = jmp, r = r, nmis = nmis,
       ro = ro, mdpgrp = mdpgrp, mdpgst = mdpgst, mobs = mobs,
       mobsst = mobsst, nmobs = nmobs, ncells = ncells, ngrp
       = ngrp, npatt = npatt, grouped = grouped, sj = sj)
}
"prelim.mix"<-
function(x, p)
{
# get dimensions of x, separate into w and z
  n <- nrow(x)
  p <- as.integer(p)
  q <- as.integer(ncol(x) - p)
  w <- x[, (1:p)]
  storage.mode(w) <- "integer"
  z <- x[, (p + 1):(p + q)]
  storage.mode(z) <- "single"		# store names
  {
    if(!is.null(dimnames(x)[[2]])) {
      wnames <- dimnames(x)[[2]][1:p]
      znames <- dimnames(x)[[2]][(p + 1):(p + q)]
    }
      else {
	wnames <- NULL
	znames <- NULL
      }
  }
  rnames <- dimnames(x)[[1]]
  w <- matrix(w, n, p)
  z <- matrix(z, n, q)			# get dimensions of contingency table
  d <- as.integer(numeric(p))
  for(j in 1:p) {
    d[j] <- as.integer(max(w[!is.na(w[, j]), j]))
  }
# missingness indicators for z
  rz <- 1 * is.na(z)
  nmisz <- as.integer(apply(rz, 2, sum))
  mdpz <- as.integer((rz %*% (2^((1:q) - 1))) + 1)
  rz <- 1 - rz
  storage.mode(rz) <- "integer"	
# get missing data patterns for w
  rw <- 1 * is.na(w)
  nmisw <- as.integer(apply(rw, 2, sum))
  nmis <- c(nmisw, nmisz)
  names(nmis) <- c(wnames, znames)
  mdpw <- as.integer((rw %*% (2^((1:p) - 1))) + 1)
  rw <- 1 - rw
  storage.mode(rw) <- "integer"	
# calculate the known part of the cell number for rows of w
  cumd <- as.integer(round(exp(cumsum(log(d)))))
  mobs <- 1 + ((.na.to.icode(w, 1) - 1) * rw) %*% (cumd/d)	
# do row sort 
  ro <- order(mdpz, mdpw, mobs)
  w <- matrix(w[ro,  ], n, p)
  mdpw <- mdpw[ro]
  mobs <- mobs[ro]
  rw <- matrix(rw[ro,  ], n, p)
  z <- matrix(z[ro,  ], n, q)
  mdpz <- mdpz[ro]
  rz <- matrix(rz[ro,  ], n, q)
  ro <- order(ro)			# compress missing data patterns
  mdpzst <- as.integer(seq(along = mdpz)[!duplicated(mdpz)])
  mdpz <- unique(mdpz)
  npattz <- length(mdpz)
  mdpzfin <- as.integer(c(mdpzst[2:npattz] - 1, n))
  if(npattz == 1)
    mdpzfin <- n
  mdpzgrp <- numeric()
  mdpwst <- numeric()
  mdpwtmp <- numeric()
  for(i in 1:npattz) {
    tmp <- mdpw[mdpzst[i]:mdpzfin[i]]
    mdpwst <- c(mdpwst, seq(along = tmp)[!duplicated(tmp)] +
		mdpzst[i] - 1)
    tmp <- unique(tmp)
    mdpzgrp <- c(mdpzgrp, length(tmp))
    mdpwtmp <- c(mdpwtmp, tmp)
  }
  mdpw <- as.integer(mdpwtmp)
  npattw <- length(mdpw)
  storage.mode(mdpzgrp) <- "integer"
  storage.mode(mdpwst) <- "integer"
  mdpwfin <- as.integer(c(mdpwst[2:npattw] - 1, n))
  mdpwgrp <- numeric()
  mobsst <- numeric()
  mobstmp <- numeric()
  for(i in 1:npattw) {
    tmp <- mobs[mdpwst[i]:mdpwfin[i]]
    mobsst <- c(mobsst, seq(along = tmp)[!duplicated(tmp)] +
		mdpwst[i] - 1)
    tmp <- unique(tmp)
    mdpwgrp <- c(mdpwgrp, length(tmp))
    mobstmp <- c(mobstmp, tmp)
  }
  mobs <- as.integer(mobstmp)
  ngrp <- length(mobs)
  storage.mode(mdpwgrp) <- "integer"
  storage.mode(mobsst) <- "integer"	
# create r-matrix for display purposes
  r <- cbind(rw, rz)[mdpwst,  ]
  tmp <- as.character(c(mdpwst[2:npattw], n + 1) - mdpwst)
  dimnames(r) <- list(tmp, c(wnames, znames))
  ncells <- cumd[p]
  jmp <- as.integer(cumd/d)
  rz <- matrix(rz[mdpzst,  ], npattz, ncol(rz))
  rw <- rw[mdpwst,  ]			# form matrix of packed storage indices
  npsi <- as.integer((q * (q + 1))/2)
  psi <- .Fortran("mkpsi",
		  as.integer(q - 1),
		  matrix(as.integer(0), q, q))[[2]]	
# center and scale the columns of z
  mvcode <- as.single(max(z[!is.na(z)]) + 1000)
  z <- .na.to.snglcode(z, mvcode)
  tmp <- .Fortran("ctrsc",
		  z,
		  n,
		  q,
		  numeric(q),
		  numeric(q),
		  mvcode)
  z <- tmp[[1]]
  xbar <- tmp[[4]]
  sdv <- tmp[[5]]
  z <- .code.to.na(z, mvcode)		# return list
  nmobs <- as.integer(c(mobsst[-1], n + 1) - mobsst)
  list(w = w, n = n, p = p, d = d, jmp = jmp, z = z, q = q, r = 
       r, rz = rz, rw = rw, nmis = nmis, ro = ro, mdpzgrp = 
       mdpzgrp, mdpwgrp = mdpwgrp, mobs = mobs, mobsst = 
       mobsst, nmobs = nmobs, ncells = ncells, ngrp = ngrp, 
       npattz = npattz, npattw = npattw, rnames = rnames, 
       npsi = npsi, psi = psi, xbar = xbar, sdv = sdv, wnames
       = wnames, znames = znames, rnames = rnames)
}
"prelim.norm"<-
function(x)
{
# get dimensions of x
  if(is.vector(x)) x <- matrix(x, length(x), 1)
  n <- nrow(x)
  p <- ncol(x)
  storage.mode(x) <- "single"		# find missingness patterns
  r <- 1 * is.na(x)
  nmis <- as.integer(apply(r, 2, sum))
  names(nmis) <- dimnames(x)[[2]]	
# index the missing data patterns
  mdp <- as.integer((r %*% (2^((1:ncol(x)) - 1))) + 1)	
# do row sort
  ro <- order(mdp)
  x <- as.matrix(x[ro,  ], n, p)
  mdp <- mdp[ro]
  r <- as.matrix(r[ro,  ], n, p)
  ro <- order(ro)			# compress missing data patterns
  mdpst <- as.integer(seq(along = mdp)[!duplicated(mdp)])
  mdp <- unique(mdp)
  npatt <- length(mdpst)		# create r-matrix for display purposes
  r <- 1 - r
  r <- matrix(r[mdpst,  ], npatt, p)
  if(npatt == 1)
    tmp <- format(n)
  if(npatt > 1)
    tmp <- format(c(mdpst[2:npatt], n + 1) - mdpst)
  dimnames(r) <- list(tmp, dimnames(x)[[2]])
  storage.mode(r) <- "integer"	
# center and scale the columns of x
  if(sum(is.na(x)) < length(x)) {
    mvcode <- as.single(max(x[!is.na(x)]) + 1000)
    x <- .na.to.snglcode(x, mvcode)
    tmp <- .Fortran("ctrsc",
		    x,
		    n,
		    p,
		    numeric(p),
		    numeric(p),
		    mvcode)
    x <- tmp[[1]]
    xbar <- tmp[[4]]
    sdv <- tmp[[5]]
    x <- .code.to.na(x, mvcode)
  }
  if(sum(is.na(x)) == length(x)) {
    xbar <- rep(0, p)
    sdv <- rep(1, p)
  }
# form matrix of packed storage indices
  d <- as.integer((2 + 3 * p + p^2)/2)
  psi <- .Fortran("mkpsi",
		  p,
		  matrix(as.integer(0), p + 1, p + 1))[[2]]	
# other bookkeeping quantities
  if(npatt > 1)
    nmdp <- as.integer(c(mdpst[-1], n + 1) - mdpst)
  if(npatt == 1)
    nmdp <- n
  sj <- .Fortran("sjn",
		 p,
		 npatt,
		 r,
		 integer(p))[[4]]
  nmon <- .Fortran("nmons",
		   p,
		   npatt,
		   r,
		   nmdp,
		   sj,
		   integer(p))[[6]]
  last <- .Fortran("lasts",
		   p,
		   npatt,
		   sj,
		   integer(npatt))[[4]]
  tmp <- .Fortran("layers",
		  p,
		  sj,
		  integer(p),
		  integer(1))
  layer <- tmp[[3]]
  nlayer <- tmp[[4]]			# return list
  list(x = x, n = n, p = p, r = r, nmis = nmis, ro = ro, mdpst
       = mdpst, nmdp = nmdp, npatt = npatt, xbar = xbar, sdv
       = sdv, d = d, psi = psi, sj = sj, nmon = nmon, last
       = last, layer = layer, nlayer = nlayer)
}
"rngseed"<-
function(seed)
{
  seed <- as.integer(seed)
  if(seed <= 0)
    stop("Seed must be a positive integer")
  tmp <- .Fortran("rngs", seed)
  invisible()
}
.First.lib <- function(library, section) 
{
  if(version$major == 3) {
    if(version$minor < 2)
      stop("This library needs S-Plus 3.2 or later")
    else sec <- "sp3"
  }
  if(version$major == 4) sec <- "sp4"
  dyn.load(paste(library, section, sec, 'AIMD.obj', sep='/'))
  invisible()
}
